#! /usr/bin/python3

import sys
import json
import datetime
import zoneinfo
import urllib.request
import psycopg2

dsn = 'dbname=investment-game user=investment-game'

lookback_days = 7


def daterange(start, end):
    return (
        datetime.date.fromordinal(o)
        for o in range(start.toordinal(), end.toordinal() + 1)
    )


asset_price_insert_query = """
        INSERT INTO stocks_assetprice (
            currency,
            closing_price,
            closing_time,
            closing_date,
            timezone_name,
            asset_id
        ) VALUES (%s, %s, %s, %s, %s, %s)
    """


def main():
    pg = psycopg2.connect(dsn)

    with pg.cursor() as cursor:
        current_timestamp = int(datetime.datetime.now().timestamp())
        today = datetime.date.today()
        lookback_date = today - datetime.timedelta(days=lookback_days)

        # For each ticker, determine the interval(s) in which it is active
        # in a portfolio revision during the past lookback_days days:
        # GME       [-------------]                 [----------------
        # BBBY              [---------------------------]
        # TSLA
        # SPCE      ------------------------------------------------]

        cursor.execute(
            """
            SELECT DISTINCT
                a.id,
                a.ticker,
                GREATEST(pfr.created_at::DATE, %(lookback_date)s)
                    AS start_date,
                LEAST(g.end_date, CURRENT_DATE)
            FROM game_game g
                JOIN game_participation p ON p.game_id = g.id
                JOIN game_portfolio pf ON pf.participation_id = p.id
                JOIN game_portfoliorevision pfr ON pfr.portfolio_id = pf.id
                JOIN game_portfolioasset pfa
                    ON pfa.portfolio_revision_id = pfr.id
                JOIN stocks_asset a ON a.id = pfa.asset_id
                WHERE g.start_date <= CURRENT_DATE
                    AND g.end_date >= %(lookback_date)s
            """,
            {'lookback_date': lookback_date},
        )

        # Hypothetical query results:
        # 1  GME   2021-04-03  2021-04-10
        # 1  GME   2021-04-15  2021-04-20
        # 3  BBBY  2021-04-07  2021-04-12
        # 7  SPCE  2021-04-03  2021-04-20

        results = list(cursor.fetchall())
        ticker_to_asset_id = {ticker: pk for pk, ticker, start, end in results}

        # Convert the intervals to sets of dates:
        asset_wanted_dates = {pk: set() for pk in ticker_to_asset_id.values()}
        for pk, ticker, start, end in results:
            asset_wanted_dates[pk].update(daterange(start, end))

        # The list of all asset ids that are in active use:
        wanted_asset_ids = tuple(
            pk for ticker, pk in ticker_to_asset_id.items()
            if asset_wanted_dates[pk]
        )

        if not wanted_asset_ids:
            # Do not continue if there are no active games/stocks:
            # the SQL ‘IN’ expression does not like empty lists.
            return

        # Now see which prices we retrieved previously:
        cursor.execute(
            """
                SELECT ap.asset_id, ap.closing_date
                    FROM stocks_assetprice ap
                    WHERE ap.asset_id IN %s
                    AND ap.closing_date >= %s
            """,
            (wanted_asset_ids, lookback_date),
        )

        # asset_old_dates will contain the dates for which we already knew
        # the value.
        asset_old_dates = {pk: set() for pk in ticker_to_asset_id.values()}
        for pk, closing_date in cursor.fetchall():
            asset_wanted_dates[pk].discard(closing_date)
            asset_old_dates[pk].add(closing_date)

        # The set of dates for which at least one asset does not have a value:
        wanted_dates = set().union(*asset_wanted_dates.values())
        if not wanted_dates:
            return
        first_wanted_date = min(wanted_dates)
        first_wanted_days = (today - first_wanted_date).days

        tickers = ','.join(
            ticker for ticker, pk in ticker_to_asset_id.items()
            if asset_wanted_dates[pk]
        )

        url = (
            "https://query1.finance.yahoo.com/v7/finance/spark"
            f"?symbols={tickers}&range={first_wanted_days}d&interval=1d"
            # "&indicators=adjclose&includeAdjustedClose=true&events=div|split"
        )

        with urllib.request.urlopen(url) as fh:
            yahoo_json = json.load(fh)

        # print(url)
        # with open('yahoo.json') as fh:
        #    yahoo_json = json.load(fh)

        for result in yahoo_json['spark']['result']:
            for response in result['response']:
                meta = response['meta']
                currency = meta['currency']
                timezone_name = meta['exchangeTimezoneName']
                timezone_info = zoneinfo.ZoneInfo(timezone_name)
                ticker = meta['symbol']
                try:
                    asset_id = ticker_to_asset_id[ticker]
                except KeyError:
                    # Apparently we got a different ticket than we asked for?
                    print(
                        f"Output listed a ticker ({ticker}) we didn't ask for",
                        file=sys.stderr,
                    )
                    continue

                old_dates = asset_old_dates[asset_id]

                # Keep a set of dates that are missing (typically weekend days)
                # so we can add dummy entries (with a NULL price) at the end.
                # We'll start with all dates we want a price for, and eliminate
                # any dates we get from the API further down.
                pending_dates = asset_wanted_dates[asset_id].copy()
                # Never try to fake an entry for today:
                pending_dates.discard(today)

                trading_period = meta['currentTradingPeriod']['regular']
                trading_start = trading_period['start']
                trading_end = trading_period['end']
                trading_duration = trading_end - trading_start

                close_quotes = response['indicators']['quote'][0]['close']
                for i, timestamp in enumerate(response['timestamp']):
                    # It's important to use the current time zone info here so
                    # we can extract the date properly. Timestamps may be in
                    # different dates depending on the time zone, after all.
                    closing_time = datetime.datetime.fromtimestamp(
                        timestamp + trading_duration,
                        tz=timezone_info,
                    )

                    closing_date = closing_time.date()
                    pending_dates.discard(closing_date)

                    # Trading is still in progress.
                    if trading_start <= current_timestamp < trading_end:
                        # Timestamp is for the current trading day and that
                        # day is not over yet, so we don't know yet what the
                        # closing price will be. Skip it.
                        continue

                    # If Yahoo! were to suddenly give us a result outside the
                    # expected date range, we might try to insert it into the
                    # database and cause a unique key violation.
                    if not (lookback_date <= closing_date <= today):
                        continue

                    # We already got one. It's very nice.
                    if closing_date in old_dates:
                        continue

                    cursor.execute(asset_price_insert_query, (
                        currency,
                        close_quotes[i],
                        closing_time,
                        closing_date,
                        timezone_name,
                        asset_id,
                    ))

                trading_end_time = datetime.datetime.fromtimestamp(
                    trading_end,
                    tz=timezone_info,
                )

                # Process the dates we expected to get from the API but didn't
                # (presumably because there was no trading that day, for
                # instance, because it was Saturday or Sunday):
                for pending_date in pending_dates:
                    # Build a fake closing time:
                    pending_time = trading_end_time.replace(
                        year=pending_date.year,
                        month=pending_date.month,
                        day=pending_date.day,
                    )
                    # Mark these days so that we don't retry fetching them
                    # every night:
                    cursor.execute(asset_price_insert_query, (
                        currency,
                        None,
                        pending_time,
                        pending_date,
                        timezone_name,
                        asset_id,
                    ))

        pg.commit()


if __name__ == '__main__':
    main()
